/*
 * fmc228.c
 *
 * Copyright 2016 VadaTech Incorporated.  All rights reserved.
 * Author: Maxwell S.
 
 * When this application is compiled under Xilinx SDK as part of the FPGA reference design,
 *  it uses the UARTLITE port as the stdin/stdout console to provide
 *  a menu interface to test the board.
 *
 * When this application is compiled under Linux, it uses PCIe/CPU Local bus(LB) interface to test the board.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "fmc228.h"

#ifdef __MICROBLAZE__
	#define PROG			""
	#define FILE_OPT		"      "
	#include "xstatus.h"
	#include "platform.h"
	#include "xparameters.h"
	#include "cli.h"
#else
	#define PROG			"fmc228"
	#define FILE_OPT		"[file]"
#endif

#ifndef __MICROBLAZE__
#if ACCESS_INTERFACE == PCIE_MMAP
	typedef struct
	{
		struct pci_access*	pci;		// PCILIB access structure
		struct pci_dev*		dev;		// PCILIB device structure
		FILE*			mem_file;	// handle to /dev/mem
		void*			base;		// mmapped base address of BAR0
	} pci_card_context_t;

	#define PCI_SUBSYSTEM_ID_POS	0x2E
	#define DESIRED_VENDOR		0xabcd
	#define AMC502_DEVICE		0x4502
	#define AMC515_DEVICE		0x4515
	#define AMC516_DEVICE		0x4516
	#define AMC517_DEVICE		0x4517
	#define AMC518_DEVICE		0x4518
	#define AMC519_DEVICE		0x4519
	#define AMC525_DEVICE		0x4525
	#define AMC527_DEVICE		0x4527
	#define DESIRED_SUB_DEVICE	0xF228
	#define DESIRED_BAR		0
	#include <errno.h>
	#include <pci/pci.h>
	#include <sys/mman.h>
	#include <linux/pci.h>
#endif
#endif

#ifndef __MICROBLAZE__
#if ACCESS_INTERFACE == P2040_LB || ACCESS_INTERFACE == PCIE_DEV_DRV || ACCESS_INTERFACE == IMX6_LB
	#include <fcntl.h>
#endif
#endif


#include "tool_common.h"
#include "spi_sdio.h"
#include "jesd204.h"
#include "hmc835.h"
#include "lmk04828.h"
#include "ad9234.h"

#define ADC_SAMPLES		4096
#define BOARD			"AMC Carrier"
#define APPLICATION		"FMC228 on AMC Carrier"
#define TOOL_VERSION		"1.0.0.0"


int			fd = 0;		/* device driver */
void*			base = 0;	/* memory mapped base address */
char *const 		prompt = ">>";

/**
 * @brief buses definitions
 */

static sdio_config_t sdio_lmk04828 = {
	.name			= "lMK04828",
	.base_addr		= LMK04828_BASE,
	.rb_shift		= 0,
	.rbpol			= 0,
	.cpol			= 0,
	.cpha			= 0,
	.rw_bit_read_pol	= 1,
	.rw_bit_pos		= 24,
	.data_pos		= 8,
	.data_width		= 8,
	.tran_width		= 24,
	.write_reg32		= write_reg32,
	.read_reg32		= read_reg32
};

static sdio_config_t sdio_hmc835 = {
	.name			= "HMC835 for ADC",
	.base_addr		= HMC835_BASE,
	.rb_shift		= 1,
	.rbpol			= 0,
	.cpol			= 1,
	.cpha			= 1,
	.rw_bit_read_pol	= 1,
	.rw_bit_pos		= 0,
	.data_pos		= 32,
	.data_width		= 24,
	.tran_width		= 32,
	.write_reg32		= write_reg32,
	.read_reg32		= read_reg32
};

static sdio_config_t sdio_ad9234_0 = {
	.name			= "AD9234_0",
	.base_addr		= AD9234_0_BASE,
	.rbpol			= 0,
	.cpol			= 0,
	.cpha			= 0,
	.rw_bit_read_pol	= 1,
	.rw_bit_pos		= 24,
	.data_pos		= 8,
	.data_width		= 8,
	.tran_width		= 24,
	.write_reg32		= write_reg32,
	.read_reg32		= read_reg32
};

static sdio_config_t sdio_ad9234_1 = {
	.name			= "AD9234_1",
	.base_addr		= AD9234_1_BASE,
	.rbpol			= 0,
	.cpol			= 0,
	.cpha			= 0,
	.rw_bit_read_pol	= 1,
	.rw_bit_pos		= 24,
	.data_pos		= 8,
	.data_width		= 8,
	.tran_width		= 24,
	.write_reg32		= write_reg32,
	.read_reg32		= read_reg32
};

/**
 * @brief serial devices definitions
 */ 
dev_with_regs_t dev_lmk =
{
	.fd			= 0,
	.base			= 0,
	.name			= "l",
	.full_name		= "LMK04828",
	.regs_dump		= lmk04828_regs,
	.regs_init		= lmk04828_regs,
	.bus_type		= SDIO,
	.pBus			= &sdio_lmk04828,
	.private_data[0]	= 0,
	.init			= dev_init,
	.write			= spi_write,
	.read			= spi_read,
	.dump			= spi_dump
};

// dev->private_data[0] format for SPI_SDIO device:
// 
// [7:0] read register
// [15:8] read address extra shift
// [23:16] extra data to write to the read register 
// [31] use write then read

// dev->private_data[1] format for SPI_SDIO device:
// 
// [15:8] write address extra shift
// [23:16] extra data to write to address field

dev_with_regs_t dev_hmc =
{
	.fd			= 0,
	.base			= 0,
	.name			= "h",
	.full_name		= "HMC835",
	.regs_dump		= hmc835_regs,
	.regs_init		= hmc835_regs,
	.bus_type		= SDIO,
	.pBus			= &sdio_hmc835,
	.private_data[0]	= 0x80000000 | 0 | 0,
	.private_data[1]	= 0x300,
	.init			= dev_init,
	.write			= spi_write,
	.read			= spi_read,
	.dump			= spi_dump
};

dev_with_regs_t dev_adc0 =
{
	.fd			= 0,
	.base			= 0,
	.name			= "a0",
	.full_name		= "AD9234 0",
	.regs_dump		= ad9234_regs,
	.regs_init		= ad9234_regs,
	.bus_type		= SDIO,
	.pBus			= &sdio_ad9234_0,
	.private_data[0]	= 0,
	.init			= spi_init,
	.write			= spi_write,
	.read			= spi_read,
	.dump			= spi_dump
};

dev_with_regs_t dev_adc1 =
{
	.fd			= 0,
	.base			= 0,
	.name			= "a1",
	.full_name		= "AD9234 1",
	.regs_dump		= ad9234_regs,
	.regs_init		= ad9234_regs,
	.bus_type		= SDIO,
	.pBus			= &sdio_ad9234_1,
	.private_data[0]	= 0,
	.init			= spi_init,
	.write			= spi_write,
	.read			= spi_read,
	.dump			= spi_dump
};

#define NUM_DEVICES		4

dev_with_regs_t* p_devices[] = {
	&dev_lmk,
	&dev_hmc,
	&dev_adc0,
	&dev_adc1
};

static void update_dev_base(int fd, void* base)
{
	int i;
	for(i = 0; i < NUM_DEVICES; i++)
	{
		p_devices[i]->fd = fd;
		p_devices[i]->base = base;		
	}		
}

#ifndef __MICROBLAZE__
#if ACCESS_INTERFACE == PCIE_MMAP
static int map_bar( FILE **mem_file, void** base, struct pci_dev *dev )
{
	printf( PROG ": BAR%d start 0x%08llX size 0x%08llX\n",
			DESIRED_BAR,
			(long long unsigned)dev->base_addr[DESIRED_BAR], (long long unsigned)dev->size[DESIRED_BAR] );

	*mem_file = fopen( "/dev/mem", "r+" );

	if( !*mem_file )
	{
		perror( PROG ": Unable to open /dev/mem" );
		return EXIT_FAILURE;
	}

	*base = mmap( NULL, dev->size[DESIRED_BAR],
				  PROT_READ|PROT_WRITE, MAP_SHARED,
				  fileno(*mem_file),
				  (off_t)dev->base_addr[DESIRED_BAR] );

	if( *base == MAP_FAILED )
	{
		fprintf( stderr, "*** ERRNO %d ***\n", errno );
		perror( PROG ": Unable to map BAR" );
		fclose( *mem_file );
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

static int unmap_bar( FILE *mem_file, void* base, struct pci_dev *dev )
{
	if( munmap( base, dev->size[DESIRED_BAR] ) < 0 )
	{
		return EXIT_FAILURE;
	}

	fclose( mem_file );

	return EXIT_SUCCESS;
}

static int pci_setup_card( pci_card_context_t* context )
{
	context->pci = pci_alloc();

	if( !context->pci )
	{
		fprintf( stderr, PROG ": Unable to allocate PCI structure\n" );
		goto out;
	}

	pci_init( context->pci );
	pci_scan_bus( context->pci );

	/* walk the list of devices and see how many belong to us */
	for( context->dev = context->pci->devices; context->dev; context->dev = context->dev->next )
	{
		pci_fill_info( context->dev, PCI_FILL_IDENT|
					   PCI_FILL_IRQ|
					   PCI_FILL_BASES|
					   PCI_FILL_ROM_BASE|
					   PCI_FILL_SIZES|
					   PCI_FILL_CLASS|
					   PCI_FILL_RESCAN );

		u16 sub_device_id = pci_read_word(context->dev, PCI_SUBSYSTEM_ID_POS);
		if( context->dev->vendor_id == DESIRED_VENDOR &&
			( context->dev->device_id == AMC502_DEVICE || 
			  context->dev->device_id == AMC515_DEVICE ||
			  context->dev->device_id == AMC516_DEVICE ||
			  context->dev->device_id == AMC517_DEVICE ||
			  context->dev->device_id == AMC518_DEVICE ||
			  context->dev->device_id == AMC519_DEVICE ||
			  context->dev->device_id == AMC525_DEVICE ||
			  context->dev->device_id == AMC527_DEVICE ) &&
			sub_device_id == DESIRED_SUB_DEVICE)
		{
			/* this is our board */
			break;
		}
		if( context->dev->vendor_id == DESIRED_VENDOR) {
			printf("found Vadatech board\n");
		}
		
	}

	if( !context->dev )
	{
		printf( PROG ": Unable to find " BOARD " board\n" );
		goto cleanup;
	}


	printf( PROG ": Found " BOARD " board at %d:%d.%d\n", context->dev->bus, context->dev->dev, context->dev->func );

	if( map_bar( &context->mem_file, &context->base, context->dev ) != EXIT_SUCCESS )
	{
		context->dev = NULL;
		goto cleanup;
	}

	goto out;

cleanup:
	pci_cleanup( context->pci );
	return EXIT_FAILURE;

out:
	return EXIT_SUCCESS;
}

static void pci_cleanup_card( pci_card_context_t* context )
{
	unmap_bar( context->mem_file, context->base, context->dev );

	pci_cleanup( context->pci );
}
#endif
#endif

static int check_jesd204_link(int fd, void* base, unsigned int jesd204_core_base)
{
	unsigned int read_data;
	read_reg32(fd, base, jesd204_core_base+JESD204_SYNC_STATUS, &read_data);

	return read_data;
}

static int sync_clk_gen(int fd, void* base)
{
	int i;

	/*
	const reg_t regs[6] =
	{
		// Normal Sync (SYSREF_MUX = 0)
		{0x139, 0x0},

		// SYNC_MODE = 1, SYNC_POL = 0, SYNC_EN = 1
		{0x143, 0x11},

		//To allow SYNC to effect dividers
		{0x144, 0x0},

		// #########  Now perform SYNC #############
		// Assert SYNC pin and deassert it

		// Disable SYNC from resetting dividers
		{0x144, 0x9F},

		// Level sensitive SYNC; enable SYNC; SYNC_MODE = 0 (Disable)
		{0x143, 0x10},

		// SYSREF_MUX = SYSREF continuous
		{0x139, 0x3}
	};
	*/

	const unsigned int addr[6] = {0x139, 0x143, 0x144, 0x144, 0x143, 0x139 };
	const unsigned int data[6] = {  0x0,  0x11,   0x0,  0x9F,  0x10,   0x3 };
	
	
	for(i=0; i<6; i++)
	{

		// Toggle SYNC pin
		if(i == 3)
		{
			// Driving the SYNC pin high will reset BUFR divider and JESD204 cores inside the FPGA too
			
			write_reg32(fd, base, GPIO_BASE + GPIO_LMK04828_SYNC_CTRL, 1);
			vt_delay(10);
			write_reg32(fd, base, GPIO_BASE + GPIO_LMK04828_SYNC_CTRL, 0);
			vt_delay(10);
		}
		
		dev_lmk.write(&dev_lmk, addr[i], data[i]);

	}

	return EXIT_SUCCESS;
}

static int usage_exit( void )
{
	fprintf( stderr, "\n" APPLICATION " Tool v%s. "
		 "Usage: " PROG " <cmd> [<opts>]\n"
		"CMDs:\n"
		"    i                             - Initialize the card to default values\n"
		"    c " FILE_OPT "                      - Capture ADC channel data\n"
		"    d <g|j0|j1>                   - Dump FPGA register contents\n"
		"    w <addr> <val>                - Write value at FPGA address\n"
		"    r <addr>                      - Read back at FPGA address\n"
		"    dd <a0|a1|h|l>                - Dump device's register contents\n"
		"    wd <a0|a1|h|l> <addr> <val>   - Write value to the device's address\n"
		"    rd <a0|a1|h|l> <addr>         - Read value from the device's address\n"
		"    rs <fpga|fp>                  - Select 10MHz ref clock input for lmk04828\n",
		TOOL_VERSION );
		
	fprintf( stderr, "Device Key:\n"
		 "\t g    = GPIO\n"
		 "\t j0   = JESD204 0\n"
		 "\t j1   = JESD204 1\n"
		 "\t a0   = AD9234 0\n"
		 "\t a1   = AD9234 1\n"
		 "\t h    = HMC835\n"
		 "\t l    = LMK04828\n"
		 );
	
#ifndef __MICROBLAZE__
	exit( EXIT_FAILURE );
#else
	return EXIT_FAILURE;
#endif
}

static int cmd_init( int argc, char* argv[] )
{
	const reg_with_name_t* reg;
	int j;
	unsigned int read_data;
	
	/* Enable on-board 100MHz VCXO */ 
	write_reg32(fd, base, GPIO_BASE + GPIO_VCXO_EN, 1);

	/* Turn on 3 USER LEDs (active low) */ 
	write_reg32(fd, base, GPIO_BASE + GPIO_USER_LED, 0);
		
	// ------------- Disable JESD204 clock input and reset the core -----------
	write_reg32(fd, base, GPIO_BASE + GPIO_JESD204_RESET, 7);

	// --------------- LMK04828 initialization	
	printf("Initializing LMK04828\n");	
	// reset the chip
	write_reg32(fd, base, GPIO_BASE + GPIO_LMK04828_RST, 1);
	vt_delay(10);
	write_reg32(fd, base, GPIO_BASE + GPIO_LMK04828_RST, 0);
	vt_delay(10);
	
	dev_lmk.init(&dev_lmk);

	// check if LMK04828 locked
	printf("Checking if the LMK04828 is locked ... ");
	for(j=0; j<50; j++)
	{
		vt_delay(500);

		int pll_locked = 0;
		dev_lmk.read(&dev_lmk, 0x182, &read_data);
		pll_locked = read_data & 2;
		
		dev_lmk.read(&dev_lmk, 0x183, &read_data);
		pll_locked &= read_data & 2;
		
		if(pll_locked == 2)
		{
			break;
		}		
	}
	
	if(j == 50)
	{
		printf("not locked after 50 ms!!!\n");
	}
	else 
	{
		printf("locked.\n");
	}
	
	//------------------ HMC835 initialization
	printf("Initializing two HMC835LP6GE chips\n");	
	// Reset the chip
	write_reg32(fd, base, GPIO_BASE + GPIO_HMC835_CTRL, 0);
	vt_delay(10);
	write_reg32(fd, base, GPIO_BASE + GPIO_HMC835_CTRL, 1);
	
	dev_hmc.init(&dev_hmc);
	
	vt_delay(10000);
			
	// check if the HMC is locked
	printf("Checking if HMC is locked ... ");
	for(j=0; j<50; j++)
	{
		vt_delay(500);

		read_reg32(fd, base, GPIO_BASE + GPIO_HMC835_LD_STATUS, & read_data);
		
		if((read_data & 1) == 1)
		{
			break;
		}
		// printf("times = %d\n", j);
	}
	
	if(j == 50)
	{
		printf("not locked after 50 ms!!!\n");
	}	
	else 
	{
		printf("locked.\n");
	}
	
		
	//-------------------- AD9234 initialization
	printf("Initializing AD9234 ADCs\n");	
	// reset the chip
	write_reg32(fd, base, GPIO_BASE + GPIO_ADC_PDWN, 3);
	vt_delay(10);
	write_reg32(fd, base, GPIO_BASE + GPIO_ADC_PDWN, 0);
	vt_delay(10*1000);
	
	dev_adc0.init(&dev_adc0);
	dev_adc1.init(&dev_adc1);
	
	// wait for ADC calibration done
	printf("Delay 300ms\n");		
	vt_delay(300*1000);

			
	// ------------- JESD204 Core initialization
	printf("Initializing JESD204 Core\n");
	
	for(reg = xilinx_jesd204_regs; reg->name; reg++)
	{
		if((reg->access_type & 2) == 2) 
		{
			write_reg32(fd, base, JESD204_0_BASE + reg->addr, reg->data);
			write_reg32(fd, base, JESD204_1_BASE + reg->addr, reg->data);
		}
	}

	sync_clk_gen(fd, base);

	vt_delay(1000);

	// Enable JESD204 clocks
	printf("Enabling JESD204 clocks\n");
	
	write_reg32(fd, base, GPIO_BASE + GPIO_JESD204_RESET, 0);

	vt_delay(1000);
	
	// Check JESD204 0 links
	printf("Checking JESD204 core 0 links ... ");	
	for(j=0; j<50; j++)
	{
		vt_delay(500);

		if(check_jesd204_link(fd, base, JESD204_0_BASE) & 1)
		{
			break;
		}

		sync_clk_gen(fd, base);
	}
	
	if(j == 50)
	{
		printf("link not established after 50ms!!!\n");	
	}
	else
	{
		printf("linked.\n");	
	}
	
	// Check JESD204 1 links
	printf("Checking JESD204 core 1 links ... ");	
	for(j=0; j<50; j++)
	{
		vt_delay(500);

		if(check_jesd204_link(fd, base, JESD204_1_BASE) & 1)
		{
			break;
		}

		sync_clk_gen(fd, base);
	}
	
	if(j == 50)
	{
		printf("link not established after 50ms!!!\n");	
	}
	else
	{
		printf("linked.\n");	
	}
		
	return EXIT_SUCCESS;
}

static int cmd_write( int argc, char* argv[] )
{
	unsigned int		write_addr;
	unsigned int		write_val;

	if( argc != 2 )
	{
		return EXIT_FAILURE;
	}

	write_addr = strtoul( argv[0], NULL, 0 );
	write_val  = strtoul( argv[1], NULL, 0 );

	printf( "Writing 0x%08X @ 0x%04X\n", write_val, write_addr );
	fflush( stdout );

	write_reg32(fd, base, write_addr, write_val);

	return EXIT_SUCCESS;
}

static int cmd_read( int argc, char* argv[] )
{
	unsigned int		read_addr;
	unsigned int		read_val;

	if( argc != 1 )
	{
		return EXIT_FAILURE;
	}

	read_addr = strtoul( argv[0], NULL, 0 );

	read_reg32(fd, base, read_addr, &read_val);

	printf( "Read value = 0x%08X @ register address 0x%04X\n", read_val, read_addr );
	fflush( stdout );

	return EXIT_SUCCESS;
}

static int cmd_write_dev( int argc, char* argv[] )
{
	int 			i;
	
	unsigned int addr;
	unsigned int write_data;
	unsigned int channel = 0;
	
	
	if(argc != 3 && argc != 4)
	{
		return EXIT_FAILURE;
	}
	
	addr = strtoul( argv[1], NULL, 0 );
	write_data = strtoul( argv[2], NULL, 0 );
	
	if(argc == 4) 
	{
		channel = strtoul( argv[3], NULL, 0);
	}
	
	for(i = 0; i < NUM_DEVICES; i++)
	{
		if(strcmp( argv[0], p_devices[i]->name ) == 0 )
		{
			if(argc == 4) {
				p_devices[i]->curr_channel = channel;
			}
			p_devices[i]->write(p_devices[i], addr, write_data);
			break;
		}		
	}		
			
	if(i == NUM_DEVICES)
	{
		fflush( stdout );
		return EXIT_FAILURE;
	}
	else
	{
		printf( "Wrote 0x%X @ 0x%X to device %s\n", write_data, addr, p_devices[i]->name );		
	}
	
	fflush( stdout );
	return EXIT_SUCCESS;
}

static int cmd_read_dev( int argc, char* argv[] )
{
	int			i;
	
	unsigned int addr;
	unsigned int read_data;
	unsigned int channel = 0;
	
	
	if(argc != 2 && argc != 3)
	{
		return EXIT_FAILURE;
	}
	
	addr = strtoul( argv[1], NULL, 0 );
	
	if(argc == 3) 
	{
		channel = strtoul( argv[2], NULL, 0);
	}
	
	for(i = 0; i < NUM_DEVICES; i++)
	{
		if(strcmp( argv[0], p_devices[i]->name ) == 0 )
		{
			if(argc == 3) 
			{
				p_devices[i]->curr_channel = channel;
			}
			p_devices[i]->read(p_devices[i], addr, &read_data);
			break;
		}		
	}		
			
	if(i == NUM_DEVICES)
	{
		fflush( stdout );
		return EXIT_FAILURE;
	}
	else
	{
		printf( "Read value = 0x%X @ register address 0x%X from device %s\n", read_data, addr, p_devices[i]->name );
	}
	
	fflush( stdout );
	
	return EXIT_SUCCESS;
}

static int cmd_dump( int argc, char* argv[] )
{
	int dumped = 0;
	
	if(argc != 1 && argc != 2)
	{
		return EXIT_FAILURE;
	}

	/* FPGA registers dump */
	if( (strcmp( argv[0], "all" ) == 0 ) || (strcmp( argv[0], "g" ) == 0 ))
	{
		reg_dump(fd, base, GPIO_BASE, "VT GPIO Core", gpio_regs);
		dumped = 1;
	}

	if( (strcmp( argv[0], "all" ) == 0 ) || (strcmp( argv[0], "j0" ) == 0 ))
	{
		reg_dump(fd, base, JESD204_0_BASE, "Xilinx JESD204 Core 0", xilinx_jesd204_regs);
		dumped = 1;
	}
	
	if( (strcmp( argv[0], "all" ) == 0 ) || (strcmp( argv[0], "j1" ) == 0 ))
	{
		reg_dump(fd, base, JESD204_1_BASE, "Xilinx JESD204 Core 1", xilinx_jesd204_regs);
		dumped = 1;
	}
	
	if(!dumped)
	{
		printf("Nothing to dump.... \n");
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

static int cmd_dump_dev( int argc, char* argv[] )
{
	int i;
	int dumped = 0;
	unsigned int channel = 0;
	
	if(argc != 1 && argc != 2)
	{
		return EXIT_FAILURE;
	}
	
	if(argc == 2) 
	{
		channel = strtoul( argv[1], NULL, 0);
	}

	/* Devices registers dump */
	for(i = 0; i < NUM_DEVICES; i++)
	{
		if(strcmp( argv[0], p_devices[i]->name ) == 0 )
		{
			printf("found device name = %s\n", p_devices[i]->name);
			if(argc == 2) 
			{
				p_devices[i]->curr_channel = channel;
			}
			p_devices[i]->dump(p_devices[i]);
			break;
		}		
	}		
	
	if(i < NUM_DEVICES)
	{
		dumped = 1;
	}
	

	if(!dumped)
	{
		printf("Nothing to dump.... \n");
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

static int cmd_ref_clk_sel( int argc, char* argv[] )
{
	unsigned int write_data = 0;
	
	if(argc != 1)
	{
		return EXIT_FAILURE;
	}
		
	if(strcmp( argv[0], "fpga" ) == 0 )
	{
		write_data = 0xA;
	}
	else if(strcmp( argv[0], "fp" ) == 0 )
	{
		write_data = 0x1A;
	}
	else
	{	
		return EXIT_FAILURE;
	}
	
	dev_lmk.write(&dev_lmk, 0x147, write_data);
		
	fflush( stdout );
	
	return EXIT_SUCCESS;
}

static int cmd_capture( int argc, char* argv[] )
{
	unsigned int read_data;
	int i;

#ifdef __MICROBLAZE__
	if(argc != 0)
#else
	if(argc != 1 &&  argc != 0)
#endif
	{
		return EXIT_FAILURE;
	}
	
	// Disable fifo write and reset fifo
	write_reg32(fd, base, GPIO_BASE + GPIO_ADC_ACQ_CTRL, 4);
	vt_delay(1);
	
	// Release fifo reset
	write_reg32(fd, base, GPIO_BASE + GPIO_ADC_ACQ_CTRL, 0);
	vt_delay(1);
	
	// Enable fifo write 
	write_reg32(fd, base, GPIO_BASE + GPIO_ADC_ACQ_CTRL, 8);
	vt_delay(1000);
	
	// Read the fifo status. the fifo should be full
	read_reg32(fd, base, GPIO_BASE + GPIO_ACQ_FIFO_STAT, &read_data);
	if((read_data & 0x10001000) != 0x10001000)
	{
		printf("ADC1 and ADC0 FIFOs are not full. FIFO status = 0x%X\n", read_data);
		goto cleanup;
	}

	printf("start reading adc fifo \n");
	
	// Save data to the file
	char* filename = "<unknown>";
	FILE* outfile;
	
	if(argc == 1) 
	{
		filename = argv[0];
		outfile = fopen(filename, "w" );
		if( !outfile )
		{
			perror( filename );
			goto cleanup;
		}
		printf( ">>> Recording to file '%s'\n", filename );	
	}
	else
	{
		outfile = stdout;
	}

	fprintf(outfile, "ADC0 A\tADC0 B\tADC1 A\tADC1 B\n");	
		
	// Now start read data. 
	unsigned short* dshort = (unsigned short*) &read_data;
	int data_i;
		
	for(i = 0; i < ADC_SAMPLES; i++)
	{

#if defined(__MICROBLAZE__) || ACCESS_INTERFACE == PCIE_MMAP || ACCESS_INTERFACE == IMX6_LB
		/* Little endian, Microblaze is set to little endian. Assuming host cpu using PCIE_MMAP is little endian */
		read_reg32(fd, base, ADC0_RD_BASE, &read_data);
		
		data_i = (int)(*(dshort+1))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\t", data_i);

		data_i = (int)(*(dshort))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\t", data_i);

		
		read_reg32(fd, base, ADC1_RD_BASE, &read_data);
		
		data_i = (int)(*(dshort+1))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\t", data_i);
	
		data_i = (int)(*(dshort))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\n", data_i);
#else 
		/* big endian: P2040_LB is big endian. Assuming host cpu using PCIE_DEV_DRV is big endian */
		read_reg32(fd, base, ADC0_RD_BASE, &read_data);
		
		data_i = (int)(*(dshort))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\t", data_i);

		data_i = (int)(*(dshort+1))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\t", data_i);

		
		read_reg32(fd, base, ADC1_RD_BASE, &read_data);
		
		data_i = (int)(*(dshort))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\t", data_i);
	
		data_i = (int)(*(dshort+1))>>4;
		if(data_i > 2047) data_i -= 4096;
		fprintf(outfile, "%d\n", data_i);
#endif
/* #if defined(__MICROBLAZE__) || ACCESS_INTERFACE == PCIE_MMAP */

		if(argc == 0 && i == 511) {
			break;	
		}
	}
			
cleanup:
	if(argc == 1 && outfile)
	{
		fclose(outfile);
	}
	return EXIT_SUCCESS;
}

#ifdef __MICROBLAZE__
int main() 
#else
int main( int argc, char* argv[] )
#endif
{
	const cmd_t*		cmd;
	const cmd_t		cmds[] =
	{
		{ "i",			cmd_init },
		{ "c",			cmd_capture },
		{ "d",			cmd_dump },
		{ "w",			cmd_write },
		{ "r",			cmd_read },
		{ "dd",			cmd_dump_dev },
		{ "wd",			cmd_write_dev },
		{ "rd",			cmd_read_dev },
		{ "rs",			cmd_ref_clk_sel },
		{ NULL,			NULL }
	};
	
	int ret_val = EXIT_FAILURE;

#if defined __MICROBLAZE__
	char input[CONFIG_SYS_CBSIZE+1];
	char* argv[20];
	int argc;

	init_platform();
	setvbuf(stdin, NULL, _IONBF, 0);

	base = (void*)XPAR_M00_AXI_BASEADDR;
	update_dev_base(fd, base);
	
	cmd_init(0, NULL);
	usage_exit();

	while (1) {

		input[0] = '\0';
		// gets_xst(prompt, input);
		cli_readline_into_buffer(prompt, input);
		
		argc = get_cmd_args(input, argv, 20);
#elif ACCESS_INTERFACE == PCIE_MMAP
	pci_card_context_t	context;
	int retval = pci_setup_card( &context );
	if( retval != EXIT_SUCCESS )
	{
		fprintf(stderr, "Failed to map PCIe address");
		return retval;
	}
	
	base = context.base;
#elif ACCESS_INTERFACE == PCIE_DEV_DRV
	
	fd = open( DEV_NAME, O_RDWR );
	
	if( fd < 0 )
	{
		perror( DEV_NAME );
		return EXIT_FAILURE;
	}

#elif ACCESS_INTERFACE == P2040_LB
	
	fd = open( get_lb_dev_name(), O_RDWR );
	
	if( fd < 0 )
	{
		perror( get_lb_dev_name() );
		return EXIT_FAILURE;
	}
	
#endif
		
	update_dev_base(fd, base);
	
	if( argc >= 2) {
		for( cmd = cmds; cmd->name; cmd++ )
		{
			if( strcmp( cmd->name, argv[1] ) == 0 )
			{
				ret_val = (cmd->func)( argc - 2, argv + 2 );
			}
		}
	}

	if(ret_val == EXIT_FAILURE) {
		usage_exit();
	}
	else {
		ret_val = EXIT_FAILURE;
	}

#if defined __MICROBLAZE__
	}

	cleanup_platform();
#elif ACCESS_INTERFACE == PCIE_MMAP
	pci_cleanup_card( &context );
	
#else
	if(fd)
	{
		close(fd);
	}
#endif
	return EXIT_SUCCESS;
}


